@@ -1,4 +1,4 @@
-use no_std_net::ToSocketAddrs;
+use no_std_net::{SocketAddr, ToSocketAddrs};
 use tinyvec::ArrayVec;
 
 /// A CoAP network socket
@@ -17,16 +17,16 @@ pub trait Socket {
   /// Send a message to the `connect`ed host
   fn send(&self, msg: &[u8]) -> nb::Result<(), Self::Error>;
 
-  /// Receive a message farom the `connect`ed host
-  fn recv(&self, buffer: &mut [u8]) -> nb::Result<usize, Self::Error>;
+  /// Pull a buffered datagram from the socket, along with the address to the sender.
+  fn recv(&self, buffer: &mut [u8]) -> nb::Result<(usize, SocketAddr), Self::Error>;
 
   /// Poll the socket for a datagram
-  fn poll(&self) -> Result<Option<ArrayVec<[u8; 1152]>>, Self::Error> {
+  fn poll(&self) -> Result<Option<(ArrayVec<[u8; 1152]>, SocketAddr)>, Self::Error> {
     let mut buf = [0u8; 1152];
     let recvd = self.recv(&mut buf);
 
     match recvd {
-      | Ok(n) => Ok(Some(buf[0..n].iter().copied().collect())),
+      | Ok((n, addr)) => Ok(Some((buf.into_iter().take(n).collect(), addr))),
       | Err(nb::Error::WouldBlock) => Ok(None),
       | Err(nb::Error::Other(e)) => Err(e),
     }
